      subroutine cmpt_intrps(pixels_array, hkl_min,
     +     hkl_max, interpolants, interpolants_size,
     +     t_analy, t_fftanaly, max_iq_slab,
     +     mem_cmpt, total_views , ctf_val, abs_ctf, algorithm)
C     CMPT_INTRPS
C **********************************************************************
C *                                                               *
C * Subroutine: cmpt_intrps                                       *
C * Author: Robert E. Lynch (rel@cs.purdue.edu)                   *
C * Date:   1998                                                  *
C * Used in: aamain -- The main routine                           *
C * Function: Fourier transform of a pixel frame and              *
C *           interpolation                                       *
C * Input:                                                        *
C *   hkl_min (Integer) -- The minimum index of HKL space.        *
C *   hkl_max (Integer) -- The maximum index of HKL space.        *
C *   pixels_array (Real) -- The array allocated in main routine  *
C *     and used for storing pixel frames read from file.         *
C * Output:                                                       *
C *   interpolants (Real) -- The results of interpolation.        *
C *   t_analy (Double precision) -- The time used in read files.  *
C *   t_fftanaly (Double precision) -- The time used in 2D DFT.   *
C *                                  interpolation.               *
C *   mem_cmpt (Double precision) -- The memory allocated in      *
C *                                   this subroutine.            *
C *                                                               *
C * Algorithm:                                                    *
C *   The first loop reads pixel frames from an external file or  *
C *   creats the frames of a sphere by calling subroutine         *
C *   "formframe".                                                *
C *                                                               *
C *   The second main loop processes a pixel frame in each        *
C *   iteration, Three steps will be taken for each view assigned *
C *   to this node:                                               *
C *   1. Compute Fourier coefficients of the pixels               *
C *   2. Find intersection points (R',K,L) and interpolated       *
C *      Fourier coefficient F(R',K,L) for each grid line, K and  *
C *      L constant. Keep three values: R', real{F(R',K,L)},      *
C *      imag{(F(R',K,L)} for each grid line.  Similarly keep     *
C *      values R', real(F(H,R',L)} for each grid line H and L    *
C *      constant and keep values R', real(F(H,K,R')} for each    *
C *      grid line H and K constant.                              *
C *   3. Average the values near a grid point.                    *
C *                                                               *
C *   The algorithm for computng Discrete Fourier Transform of    *
C *   pixel values is:                                            *
C *                                                               *
C *   The pixel values are rearranged and stored in the larger    *
C *   array                                                       *
C *        RFFT( 1:NRFFT_ROW, 1:NRFFT_COL )                       *
C *   where both NRFFT_ROW and NRFFT_COL are even; the extra      *
C *   spaces in RFFT are filled with zeros.  The discrete complex *
C *   Fourier coefficients of the nonzero rows of RFFT are        *
C *   computed and then the coefficients of the columns are       *
C *   computed.  By changing the phases of the coefficients, the  *
C *   origin is shifted so that the origin of the pixel values is *
C *   moved to (CENTER(1),CENTER(2)). The complex coefficients    *
C *   are stored in the array CFFT so that                        *
C *   CFFT( 1:NCFFT_ROW, 1:NCFFT_COL ) contains the coefficients  *
C *   arranged in the standard way that DFFT store values.  These *
C *   are rearranged so that interpolation is easy to carry out.  *
C *   The output array is complex:                                *
C *                                                               *
C *        FC_2d( -NFC_2D_ROW:NCF_2D_ROW, 0:NFC_2D_COL)           *
C *                                                               *
C *   The first index corresponds to k and the second to h which  *
C *   are the transformations subscripts for y and x,             *
C *   respectively                                                *
C *                                                               *
C *   The method for interpolation is:                            *
C *                                                               *
C *   Set the rotation matrices which transform the plane         *
C *   containing the projection of the particle to the (X,Y,Z)    *
C *   `standard space' -- by calling subroutine "setrotmat".      *
C *                                                               *
C *   Set the matrix which is used to determine the intersection  *
C *   of grid line in the standard 3-d space with the rotated     *
C *   place containing the Fourier coefficients of the pixels     *
C *   -- by calling subroutine "rottrans".                        *
C *                                                               *
C *   Finally, interpolation is done in subroutine                *
C *   "matrixentries"                                             *
C *                                                               *
C * Modification history:                                         *
C *   - Hong Lin, Version 0.6, Jul.6, 1998                      *
C *     Move code for Fourier transform and interpolation from  *
C *     the main routine to this subroutine, localize some      *
C *     dynamically allocated arrays                            *
C *   - Hong Lin, Version 1.0, Aug.18, 1998                     *
C *   - Hong Lin, Version 2.0, Sep.30, 1999.                    *
C *     Adding the "piecewise constant" scheme for              *
C *       interpolation.                                        *
C *     Adding the function for creating DAT and PIF file from  *
C *       internal data.                                        *
C *     Adding several object modes for internal test.          *
C *     Adding several schemes for generating orientations.     *
C *   - R.E. Lynch, Aug 2000                                           *
C *     Removed some statements which were left when the `internal_test'
C *       mode was eliminated                                          *
C *     Replaced several `do -- enddo' loops with                      *
C *       `do label ... label continue'.                               *
C *     Replaced several `omege' with `omega'.                         *
C *     Introduced new variable  interpolants_size  which is set in    *
C *       aamain  and is passed to this routine as an argument.        *
C *       The  size  can is one value when slab-interpolation is used  *
C *       and is another when the slab_interpolation is not used.  The *
C *       value of  interpolants_size  is checked in subsequent        *
C *       subroutine calls.                                            *

C *   - Yongchang Ji, Final Version, Aug.2000 -- Jan. 2002             *
C *                                                                    *
C **********************************************************************
      implicit none
      
      include 'mpif.h'
      include '../include/allprogs.inc'
      include '../include/info.inc'
      include '../include/com_01.inc'
      
C     Subroutine arguments
      integer hkl_min,hkl_max,max_iq_slab
      integer interpolants_size, algorithm
      real pixels_array(npix_row, npix_col, view_perread) 
      
#ifdef SPREC
      real  interpolants(interpolants_size)
#else
      double precision  interpolants(interpolants_size)
#endif
      
      real ctf_val( (nfc_2d_row+1)* (2*nfc_2d_col+1) )
      real abs_ctf( 0:nfc_2d_row, -nfc_2d_col:nfc_2d_col )
      double precision mem_cmpt, t_analy, t_fftanaly
      
C     The value of  interpolants_size  is checked in calls of
C     matrixentries  and  matrixentries_slab
C     =====================================================================
      
C     Local variables
      integer view, symm, i, j, total_views, max_node_eq
      integer max_count, datalen
      integer start_view, end_view,
     +     num_views, tmp_row
      integer*8 clock1, clock2, clockmax, clockrate, ticks
      
      real du, dv
      real    angles_rad(1:3),angles_deg(1:3),center(1:2),imgmagfactor,
     +     theta_ref, phi_ref, X_ref, Y_ref, Z_ref, num_annulus
      real    rot_mat(1:3,1:3),
     +     ruv_mat_1(1:3,1:2,1:3), ruv_mat_2(1:3,1:2,1:3),
     +     ruv_sup_1(3,3), ruv_sup_2(3,3)
      
C     Allocatable arrays
      real, allocatable ::  orient(:), rffta_omega(:), rffta(:), 
     +     rffta_trns(:), ico_x(:), ico_y(:), ico_z(:), 
     +     ico_x_ref(:), ico_y_ref(:), ico_z_ref(:)
      real, allocatable ::  HKL_DFT(:)
      complex, allocatable :: fc_2d(:)
      
      real omega_temp_1, omega_temp
      integer alloc_stat
      double precision :: wtime, t1, t2, time
      
      double precision :: mem_orient = 0.0D0,
     +     mem_rffta_omega = 0.0D0, mem_rffta = 0.0D0,
     +     mem_rffta_trns = 0.0D0, mem_ico_x = 0.0D0,
     +     mem_ico_y = 0.0D0, mem_ico_z = 0.0D0,
     +     mem_ico_x_ref = 0.0D0, mem_ico_y_ref = 0.0D0,
     +     mem_ico_z_ref = 0.0D0, mem_ico_x_original = 0.0D0,
     +     mem_ico_y_original = 0.0D0, mem_ico_z_original = 0.0D0,
     +     mem_fc_2d = 0.0D0, mem_fc_2d_tmp = 0.0D0,
     +     mem_HKL_DFT  = 0.0D0, mem_interp = 0.0D0,
     +     mem_interpmax = 0.0D0
      
      logical firsttime_symmetry
      
      character(len=12) :: subname = 'cmpt_intrps'
      
      call system_clock(count_max=clockmax, count_rate=clockrate)
      
C     Allocate storage for subroutine FFTANALY
      allocate(orient(7*max_views),
     +     rffta_omega(nrffta_omega),
     +     rffta(nrffta_row * nrffta_col),
     +     rffta_trns(nrffta_row * nrffta_col), 
     +     fc_2d((1 + nfc_2d_row) * (1 + 2 * nfc_2d_col)),
     +     ico_x(num_symm),
     +     ico_y(num_symm),
     +     ico_z(num_symm),
     +     ico_x_ref(num_symm),
     +     ico_y_ref(num_symm),
     +     ico_z_ref(num_symm),
     +     stat = alloc_stat)
      
      if( alloc_stat .gt. 0 ) call error_stop (progname, subname,
     +     'Error in allocating memory for orient')
      
      mem_orient      = dble(4*7*max_views)
      mem_rffta_omega = dble(4 * nrffta_omega)
      mem_rffta       = dble(4 * nrffta_row * nrffta_col)
      mem_rffta_trns  = dble(4 * nrffta_row * nrffta_col)
      mem_fc_2d       = dble(8 * (1 + nfc_2d_row)*(1 + 2 * nfc_2d_col))
      mem_ico_x       = dble(4 * num_symm)
      mem_ico_y       = dble(4 * num_symm)
      mem_ico_z       = dble(4 * num_symm)
      mem_ico_x_ref   = dble(4 * num_symm)
      mem_ico_y_ref   = dble(4 * num_symm)
      mem_ico_z_ref   = dble(4 * num_symm)
      
      mem_cmpt = mem_orient + mem_rffta_omega + 
     +     mem_rffta + mem_rffta_trns + mem_fc_2d + 
     +     mem_ico_x + mem_ico_y + mem_ico_z + 
     +     mem_ico_x_ref + mem_ico_y_ref + mem_ico_z_ref +
     +     mem_ico_x_original + mem_ico_y_original +
     +     mem_ico_z_original + mem_fc_2d_tmp 
      
C     to initialize the value of rffta_omega(nrfft_col+15)
C     that is rffta_omega(nrffta_omega)
C     Initialize the FFT parameters used to transform rows of RFFTA
C     each row has length NRFFTA_COL
      
      call vrffti( nrffta_col, rffta_omega )

C     ----------------------------------------------------------------------
      call system_clock(clock1)
      
      num_views = 0 
      
      if( iam .eq. 0 ) then
C     CALL READORIENT
C     ***************
         call readorient(orient, max_views, ort_file_name, num_views,
     +        npix_row, npix_col, oddeven)
         
         write(stdoutput,1001) iam, ort_file_name, num_views, max_views
 1001    format(/' Node = ',i2,':  Read orientation file: ',a/,
     +        12x,' Number of views     = ', i9, i9)
      endif
      
      if( num_nodes .gt. 1 ) then
C     call MPI_BCAST
C     ***************
         call mpi_bcast(num_views, 1, mpi_integer,
     +        0, mpi_comm_world, mpi_err)
         if( mpi_err .ne. 0 ) call error_stop(progname,subname,
     +        'Error in calling MPI_BCAST num_views')
         
         call mpi_bcast(orient, 7*max_views, mpi_real,
     +        0, mpi_comm_world, mpi_err)
         if( mpi_err .ne. 0 ) call error_stop(progname,subname,
     +        'Error in calling MPI_BCAST orient')
      endif
      
      call system_clock(clock2)
      ticks = clock2-clock1
      if(ticks .lt. 0) then
         ticks = clockmax + ticks
      endif
      t_analy = t_analy + dble(ticks)/dble(clockrate)
      
      total_views = total_views + num_views
      
C     calculate the radius in FFT space for 3D DFT
      num_annulus = real(nrffta_row)*PIXSIZ_OUT/res_max 
      if (num_annulus .ge. nrffta_row/2. )
     +     call error_stop(progname,subname,
     +     'Resolution is too high, should smaller than 2*PIXEL_SIZE')
      
C     we seperate all the views by view_perread, each time we read at
C     most view_perread * num_nodes pixels and deal with them
      
      if (algorithm .eq. 1) then
         max_count = 0
      else if (algorithm .eq. 3) then
         max_count = ((hkl_max+1)**2) * 2 * 3
      endif
      
      allocate(HKL_DFT(max_count * 6), stat = alloc_stat)
      mem_HKL_DFT     = dble( 4 * max_count * 6)
      if(alloc_stat .gt. 0 ) then
         call error_stop (progname, subname,
     +        'Error in allocating memory for HKL_DFT')
      endif
      mem_cmpt = mem_cmpt + mem_HKL_DFT 
      
      DO j = 0, (num_views-1)/(view_perread*num_nodes)
         call system_clock(clock1)
         
         start_view = 1 + j * view_perread*num_nodes
         end_view   = start_view + view_perread*num_nodes - 1
         if (end_view .gt. num_views) end_view = num_views
         
C     the pix_file_name has been opened in aamain
C     Read the current end_view-start_view+1 pixel frames.
         
         call set_indices(num_nodes, start_view, end_view,
     +        frst_lst_view, max_node_eq )
C     CALL READ_FILES
C     ***************
C     processor 0 read pixels_array, then send to processor i
         do i = num_nodes - 1 , 0 , -1
            if (frst_lst_view(2, i) .GE. frst_lst_view(1, i)) then
               call read_files(pixels_array, frst_lst_view(1,i),
     +              frst_lst_view(2,i), i, orient)
            endif
         enddo
         
         call system_clock(clock2)
         ticks = clock2-clock1
         if(ticks .lt. 0) then
            ticks = clockmax + ticks
         endif
         t_analy = t_analy + dble(ticks)/dble(clockrate)
         
Cji      do with each view and pixel in  per read

         start_view = frst_lst_view(1,iam)
         end_view   = frst_lst_view(2,iam)
         datalen    = 0
        
         do view = start_view, end_view
C     
C     Get the orientation angles and center of next pixel frame
C     Also changes OMEGA from the range (0,360) deg to (-90,270) deg.
C     Because the orient used by BIO not consistent with defination 
C     CALL GETORIENT
C     **************
            orient((view-1)*7+4) =  360 - orient((view-1)*7+4) - 90
            call getorient( orient, max_views, view,
     +           angles_rad, angles_deg, center, imgmagfactor)
C     adjust the imgmagfactor according to its own pixel_size
            imgmagfactor = imgmagfactor * PIXSIZ_OUT / angstrom_per_pix

C     The algorithm for calculating image orientations when
C     icosahedral symmetry is imposed. For one orientation
C     in the asymmetric unit, 59 orientations are to be 
C     calculated from which the view of the object is the same.

            if( symm_code .eq. 532 ) then
C     CALL ICO_EM4IMR
C     ***************
               call ICO_EM4IMR (num_symm, ICO_X, ICO_Y, ICO_Z,
     +              max_views, VIEW, orient, deg_to_rad, rad_to_deg)
               
               theta_ref = 90.
               phi_ref   = 0.
               if( abs( theta_ref-ICO_X(1) ) .le. 1. ) theta_ref = 85.
               if( abs( phi_ref-ICO_Y(1)   ) .le. 5. ) phi_ref = 20.
C     CALL ICO_EM4IMR_vector
C     **********************
               call ICO_EM4IMR_vector (num_symm, 
     +              ICO_X_ref, ICO_Y_ref, ICO_Z_ref, 
     +              theta_ref, phi_ref, deg_to_rad)
               
               firsttime_symmetry = .true.
 40            do i = 1, num_symm
                  angles_rad(1) = ICO_X(i) * deg_to_rad
                  angles_rad(2) = ICO_Y(i) * deg_to_rad
                  angles_rad(3) = ICO_Z(i) * deg_to_rad
                  
C     THE STATEMENT REPLACES THE VALUES IN THE ARRAY rot_mat WITH THE
C     VALUES OF THE INVERSE OF THE MATRIX IN rot_mat.  
C     CALL SETROTMAT
C     **************
                  call setrotmat( angles_rad, rot_mat)
                  
                  X_ref = rot_mat(1,1) * ICO_X_ref(i) +
     +                 rot_mat(1,2) * ICO_Y_ref(i) +
     +                 rot_mat(1,3) * ICO_Z_ref(i)
                  Y_ref = rot_mat(2,1) * ICO_X_ref(i) +
     +                 rot_mat(2,2) * ICO_Y_ref(i) +
     +                 rot_mat(2,3) * ICO_Z_ref(i)
                  if (abs(X_ref).lt.0.001 .and.
     +                 abs(Y_ref).lt.0.001) 
     +                 call error_stop(progname, subname,
     +                 'Error in calculating Omegas')
                  
                  omega_temp = atan2(Y_ref, X_ref)
                  if (i.eq.1) omega_temp_1 = omega_temp
                  if (firsttime_symmetry)
     +                 ICO_Z(i) = mod(orient((view-1)*7+4) + 
     +                 (omega_temp - omega_temp_1)*rad_to_deg + 720., 
     +                 360.)
                  
               enddo
               if (firsttime_symmetry) then
                  firsttime_symmetry = .false.
                  go to 40
               endif
               
            elseif (symm_code .eq. 8) then
C     CALL EIGHT_SYMMETRY
C     *******************
               call eight_symmetry(ICO_X, ICO_Y, ICO_Z, angles_deg)
               
            elseif (symm_code.eq.5 .or. symm_code.eq.52 .or.
     1              symm_code.eq.2 .or. symm_code.eq.222 .or.
     2              symm_code.eq.3 .or. symm_code.eq.32 .or.
     3              symm_code.eq.6 .or. symm_code.eq.622 .or.
     4              symm_code.eq.7 .or. symm_code.eq.72 .or.
     5              symm_code.eq.8 .or. symm_code.eq.822) then
C     CALL DIHEDRAL_SYMMETRY
C     **********************
               call dihedral_symmetry(ICO_X, ICO_Y, ICO_Z, angles_deg, 
     +              symm_code, num_symm)
               
            elseif (symm_code .eq. 1) then
               ICO_X(1) = angles_deg(1)
               ICO_Y(1) = angles_deg(2)
               ICO_Z(1) = angles_deg(3)
            else
               call error_stop(progname, subname,
     +              'Error: unsupported symmetry')
            endif
            
C------------------------------------------------------------
            
            call system_clock(clock1)
            
C     Compute Discrete Fourier Transform of pixel values:
C     CALL FFTANALY
C     *************
            call fft_2dfft( pixels_array(1, 1, view-start_view+1), 
     +           npix_row, npix_col, center,
     +           nrffta_omega, rffta_omega, nrffta_row, 
     +           nrffta_col, nfc_2d_row, 
     +           nfc_2d_col, fc_2d )
            du = center(1) - real(int(center(1)))
            dv = center(2) - real(int(center(2)))
            call shift_fft(fc_2d, nfc_2d_row, nfc_2d_col, du, dv)
            
C     ji ctf correction
            IF(CTFMODE.GT.0 .and. CTFMODE.LE.4)  then
               do tmp_row = 1, ( 1 + nfc_2d_row ) * (1 + 2 * nfc_2d_col)
                  fc_2d(tmp_row) = fc_2d(tmp_row) * ctf_val(tmp_row)
               enddo
            endif
            
            call system_clock(clock2)
            ticks = clock2-clock1
            if(ticks .lt. 0) then
               ticks = clockmax + ticks
            endif
            t_fftanaly = t_fftanaly + dble(ticks)/dble(clockrate)
            
C------------------------------------------------------------
            
            do symm = 1, num_symm
               angles_rad(1) = ico_x(symm) * deg_to_rad
               angles_rad(2) = ico_y(symm) * deg_to_rad
               angles_rad(3) = ico_z(symm) * deg_to_rad
               
C     Set the rotation matrices which transform the plane containing
C     the projection of the particle to the (X,Y,Z) `standard space'
C     and get the inverse rot_mat
C     CALL SETROTMAT
C     **************
               call setrotmat( angles_rad, rot_mat) 
               
C     Set the matrix which is used to determine the intersection of
C     grid line in the standard 3-d space with the rotated place
C     containing the Fourier coefficients of the pixels
C     CALL ROTTRANS
C     *************
Cji            call rottrans_1( rot_mat, ruv_mat_1, ruv_sup_1, -0.5 )
Cji	    call rottrans_1( rot_mat, ruv_mat_2, ruv_sup_2,  0.5 )
               call rottrans_1( rot_mat, ruv_mat_1, ruv_sup_1, -0.999 )
               call rottrans_1( rot_mat, ruv_mat_2, ruv_sup_2,  0.999 )

C     CALL MATRIXENTREIS_SLAB
C     ***********************
               call matrixentries_slab( nfc_2d_row, nfc_2d_col, fc_2d,
     +              hkl_min, hkl_max, interpolants,
     +              HKL_DFT, max_count, datalen,
     +              rot_mat, ruv_mat_1, ruv_mat_2, ruv_sup_1, ruv_sup_2,
     +              num_annulus, abs_ctf, imgmagfactor, mapmagfactor,
     +              algorithm)
               
               if (algorithm .eq. 3) then
                  mem_interp = 0
                  call exch_intpol(HKL_DFT, max_count, datalen,
     +                 interpolants, hkl_min, hkl_max, max_iq_slab, 
     +                 mem_interp)
                  mem_interpmax = max( mem_interpmax, mem_interp)
                  datalen = 0
               endif
               
            enddo
            
         enddo
                  
         if(algorithm .eq. 3) then
            do view = 1, max_node_eq - (end_view - start_view + 1)
               do symm = 1, num_symm
                  mem_interp = 0
                  call exch_intpol(HKL_DFT, max_count, datalen,
     +                 interpolants, hkl_min, hkl_max, max_iq_slab, 
     +                 mem_interp)
                  mem_interpmax = max( mem_interpmax, mem_interp)
                  datalen = 0
               enddo
            enddo
         endif
         
      enddo
      
C     Free memory allocated in this subroutine.

      deallocate(orient, rffta_omega, rffta, rffta_trns, fc_2d,
     +     ico_x, ico_y, ico_z, ico_x_ref, ico_y_ref, ico_z_ref)
      
      return
      end subroutine cmpt_intrps
