C  ****************************************************************************
C  *                                                                          *
C  *                  *****     SUBROUTINE CTF_SCALE    *****                 *
C  *                                                                          *
C  *    SCALE (MULTIPLY or DIVIDE) IMAGE FFT BY CTF.                          *
C  *                                                                          *
C  *    THE FORMULA USED depends on the value of FILTER:                      *
C  *                                                                          *
C  *    FILTER   FORMULA                                                      *
C  *    --------------------------------------------------------------------  *
C  *      0      FFTnew = FFTold*ABS(CTF*EXP(-TFAC)) [Multiplying by CTF]     *
C  *      1      FILTER 4 (before first peak)+ FILTER 2 (after first peak)    *
C  *      2      FFTnew = (FFTold*ABS(CTF)*exp(TFAC))/(CTF**2 +               *
C  *                                   FILT_FAC*(1-ABS(CTF))                  *
C  *      3      FFTnew = (FFTold*ABS(CTF)*exp(TFAC))/(CTF**2 + FILT_FAC)     *
C  *      4      FFTnew = (FFTold*exp(TFAC))/(ABS(CTF) + FILT_FAC*(1-ABS(CTF))*
C  *      5      FFTnew = (FFTold*exp(TFAC))/(ABS(CTF) + FILT_FAC)            *
C  *                                                                          *
C  *    FILT_FAC = FACTOR TO MAKE SURE NOISE IS NOT MAGNIFIED                 *
C  *               TOO MUCH NEAR THE NODES OF THE CTF.                        *
C  *    PHI     = ANGLE (IN RADIANS) SPECIFYING THE DIRECTION OF FOCUS_MAJ    *
C  *               RELATIVE TO THE X-AXIS OF THE DIGITIZED IMAGE.             *
C  *    PSI     = ANGLE (IN RADIANS) OF A PARTICULAR FOURIER COMPONENT WHOSE  *
C  *               DEFOCUS VALUE IS TO BE CALCULATED.                         *
C  *    TFAC = TEMPERATURE FACTOR, SPECIFIED IN ANGSTROMS SQUARED, IS         *
C  *               USED TO SHARPEN UP HIGHER RESOUTION FEATURES.  VALUES      *
C  *               FOR CRYOTEM DATA TYPICALLY RANGE BETWEEN 100-500 A**2.     *
C  *               THIS TEMPERATURE FACTOR IS APPLIED AS AN INVERSE T_FAC.    *
C  *    CTFMODE = 0 DO NOTHING!                                               *
C  *            = 1 CORRECT PHASES AND AMPLITUDES                             *
C  *            = 2 FLIP PHASES WHERE NECESSARY; LEAVE AMPLITUDES UNCHANGED   *
C  *            = 3 ADJUST AMPLITUDES; LEAVE PHASES UNCHANGED                 *
C  *                                                                          *
C  *    AUGUST 1999 PURDUE UNIVERSITY AlphaStation 500/400, T. BAKER.         *
C  *    MODIFIED 12-AUG-1999 BY TSB (Fixed XTEST,YTEST problems).             *
C  *    MODIFIED  2-MAY-2000 BY TSB (Added CTF_MODE switch and CTF decay).    *
C  *    MODIFIED 20-JUN-2002 BY TSB/XZ/RWA (Implementing mult filters)        *
C  *    LAST MODIFIED 12-APR-2006 BY RSS (Merged ctf_divide and _multiply     *
C  ****************************************************************************
      
      SUBROUTINE CTF_SCALE (CTFMODE, FILTER, FFT, IDIM1, IDIM2, UNITS, 
     1     PIXSIZE, PI, RADIAN, WAVELENGTH, FOCUS_MAJ, FOCUS_MIN,
     2     ANG_MAJ, AMP_FAC, Cs, FILT_FAC, TFAC, FFT_FULL, 
     3     Cc, BETA, dE, VOLTS)
      
      IMPLICIT NONE
      
      INTEGER*4, intent(in) :: CTFMODE, IDIM1, IDIM2, FILTER, UNITS
      REAL*4, intent(inout) :: FFT(2*idim1,idim2)
      real*4, intent(in)    :: PIXSIZE
      real*4, intent(in)    :: PI, RADIAN, WAVELENGTH, AMP_FAC, Cs
      real*4, intent(in)    :: Cc, BETA, dE, VOLTS, FILT_FAC(10)
      real*4, intent(in)    :: FOCUS_MAJ, FOCUS_MIN, ANG_MAJ, TFAC
      LOGICAL, intent(in)   :: FFT_FULL
      
      INTEGER*4 IX, IY, IY2
      INTEGER*4 XTEST, YTEST
      REAL*4    YSQ, CTF, D_STAR, PSI, PHI, C, T
      REAL*4    FOCUS
      REAL*4    X, Y, TMP, FACTOR
      REAL*4    FFT_RAD, FFT_RAD_TEST
      LOGICAL   ASTIG
      
      IF(CTFMODE <= 0 .OR. CTFMODE > 3) RETURN
      
      PHI = ANG_MAJ * RADIAN    ! ANGLE OF FOCUS_MAJ WRT X-DIRECTION
      
      IF(UNITS == 2) THEN
         TMP = 1.0/(IDIM1*PIXSIZE) ! ALREADY IN NM UNITS
      ELSE
         TMP = 10.0/(IDIM1*PIXSIZE) ! CONVERT TO NM UNITS
      ENDIF
      
      IF(FOCUS_MAJ == FOCUS_MIN) THEN
         ASTIG = .FALSE.
         FOCUS = FOCUS_MAJ
      ELSE
         ASTIG = .TRUE.
      ENDIF
      
      XTEST  = IDIM1/2
      YTEST  = IDIM2/2
      
      IF(FFT_FULL) THEN
         IY2 = IDIM2
      ELSE
         IY2 = YTEST + 1
      ENDIF
      
      DO IY = 1,IY2             ! PROCESS FFT A ROW AT A TIME
         IF(IY > YTEST) THEN
            Y = FLOAT(IY-IDIM2-1)
         ELSE
            Y = FLOAT(IY-1)
         ENDIF
         YSQ = Y*Y
         DO IX = 1,IDIM1        ! PROCESS CURRENT ROW
            IF(IX <= XTEST) THEN
               X = FLOAT(IX-1)  ! RIGHT HALF OF FFT
            ELSE
               X = FLOAT(IX-IDIM1-1) ! LEFT HALF OF FFT
            ENDIF
            D_STAR = SQRT(YSQ + X*X) * TMP
            IF(ABS(X) > 0.9) THEN ! TRAP FOR FFT Y-AXIS DATA
               PSI = ATAN2 (Y,X) ! ATAN2 GIVES ANGLE IN RADIANS
            ELSE
               IF(Y >= 0.0) THEN
                  PSI =  PI/2.0
               ELSE
                  PSI = -PI/2.0
               ENDIF
            ENDIF
            IF(ASTIG) CALL CTF_ASTIG (FOCUS,FOCUS_MAJ,FOCUS_MIN,PSI,PHI)
            C = CTF (D_STAR, PI, WAVELENGTH, FOCUS, AMP_FAC, Cs, Cc,
     1           BETA, dE, VOLTS)
            
            IF(FILTER == 0) THEN ! Multiplying by CTF
               IF(TFAC == 0.0) THEN
                  T = 1.0
               ELSE
                  T = EXP( -((TFAC/400.0) * D_STAR**2) )
               ENDIF
               FACTOR = ABS(T*C)
            ELSE IF(FILTER == 1) THEN ! Mixed filter 4 and 2
               IF(TFAC == 0.0) THEN
                  T = 1.0
               ELSE
                  T = EXP ((TFAC/400.0) * D_STAR**2)
               ENDIF
               CALL CTF_FIRSTPEAK_PFT (FOCUS, WAVELENGTH, FFT_RAD, 
     1              PIXSIZE, IDIM1,Cs, PI, units, amp_fac)
               FFT_RAD_TEST = D_STAR / TMP
               IF (FFT_RAD_TEST <= FFT_RAD) THEN
                  FACTOR = T/(ABS(C) + FILT_FAC(1)*(1.0-ABS(C)))
               ELSE
                  FACTOR = (T*abs(C))/(C*C + FILT_FAC(2)*(1.0-abs(C)))
               ENDIF
            ELSE IF(FILTER == 2) THEN ! "Misha+Wei"
               IF(TFAC == 0.0) THEN
                  T = 1.0
               ELSE
                  T = EXP ((TFAC/400.0) * D_STAR**2)
               ENDIF
               FACTOR = (T*ABS(C))/(C*C + filt_fac(1)*(1.0-ABS(C))) 
            ELSE IF(FILTER == 3) THEN ! "Misha" filter
               IF(TFAC == 0.0) THEN
                  T = 1.0
               ELSE
                  T = EXP ((TFAC/400.0) * D_STAR**2)
               ENDIF
               FACTOR = (T*ABS(C))/(C*C + filt_fac(1)) 
            ELSE IF(FILTER == 4) THEN ! "Wei" filter
               IF(TFAC == 0.0) THEN
                  T = 1.0
               ELSE
                  T = EXP ((TFAC/400.0) * D_STAR**2)
               ENDIF
               FACTOR = T/(ABS(C) + filt_fac(1)*(1.0-ABS(C))) 
            ELSE IF(FILTER == 5) THEN ! Original filter
               IF(TFAC == 0.0) THEN
                  T = 1.0
               ELSE
                  T = EXP ((TFAC/400.0) * D_STAR**2)
               ENDIF
               FACTOR = T/(ABS(C) + filt_fac(1)) 
            ENDIF
            
            IF(CTFMODE == 1) THEN ! ADJUST PHASES AND AMPLITUDES
               IF(C > 0.0) FACTOR = -FACTOR
               FFT(2*IX-1,IY)  = FACTOR * FFT(2*IX-1,IY)
               FFT(2*IX,IY) = FACTOR * FFT(2*IX,IY)
            ELSE IF(CTFMODE == 2) THEN ! ADJUST PHASES ONLY
               IF(C > 0.0) THEN
                  FFT(2*IX-1,IY)  = -FFT(2*IX-1,IY)
                  FFT(2*IX,IY) = -FFT(2*IX,IY)
               ENDIF
            ELSE IF(CTFMODE == 3) THEN ! ADJUST AMPLITUDES ONLY
               FFT(2*IX-1,IY)  = FACTOR * FFT(2*IX-1,IY)
               FFT(2*IX,IY) = FACTOR * FFT(2*IX,IY)
            ENDIF
         ENDDO
      ENDDO
      
      RETURN
      END subroutine ctf_scale
